import collections
import json
from collections.abc import Iterable
from pathlib import Path
from typing import Any, cast

import pandas as pd

from stac_generator.core.base import (
    CollectionGenerator,
    ItemGenerator,
    SourceConfig,
    StacCollectionConfig,
)
from stac_generator.core.base.utils import read_source_config
from stac_generator.core.point import PointGenerator
from stac_generator.core.raster import RasterGenerator
from stac_generator.core.vector import VectorGenerator

EXTENSION_MAP: dict[str, type[ItemGenerator]] = {
    "csv": PointGenerator,
    "txt": PointGenerator,
    "geotiff": RasterGenerator,
    "tiff": RasterGenerator,
    "tif": RasterGenerator,
    "zip": VectorGenerator,
    "geojson": VectorGenerator,
    "json": VectorGenerator,
    "gpkg": VectorGenerator,  # Can also contain raster data. TODO: overhaul interface
    "shp": VectorGenerator,
}


class StacGeneratorFactory:
    @staticmethod
    def register_handler(extension: str, handler: type[ItemGenerator], force: bool = False) -> None:
        if extension in EXTENSION_MAP and not force:
            raise ValueError(
                f"Handler for extension: {extension} already exists: {EXTENSION_MAP[extension]}. If this is intentional, use register_handler with force=True"
            )
        if not issubclass(handler, ItemGenerator):
            raise ValueError(
                "Registered handler must be an instance of a subclass of ItemGenerator"
            )
        EXTENSION_MAP[extension] = handler

    @staticmethod
    def get_handler(extension: str) -> type[ItemGenerator]:
        """Factory method to get ItemGenerator class based on given extension

        :param extension: file extension
        :type extension: str
        :raises ValueError: if ItemGenerator handler class for this file extension has not been registered_
        :return: handler class
        :rtype: type[ItemGenerator]
        """
        if extension not in EXTENSION_MAP:
            raise ValueError(
                f"No ItemGenerator matches extension: {extension}. Either change the extension or register a handler with the method `register_handler`"
            )
        return EXTENSION_MAP[extension]

    @staticmethod
    def get_stac_generator(
        source_configs: list[str], collection_cfg: StacCollectionConfig
    ) -> CollectionGenerator:
        handler_map = StacGeneratorFactory.match_handler(source_configs)
        handlers = [k(v) for k, v in handler_map.items()]
        return CollectionGenerator(collection_cfg, handlers)

    @staticmethod
    def generate_config_template(
        source_configs: list[str],
        dst: str,
    ) -> None:
        # Determine config type based on file extension
        if dst.endswith(".json"):
            config_type = "json"
        elif dst.endswith(".csv"):
            config_type = "csv"
        else:
            raise ValueError("Expects csv or json template")
        # Match config type with corresponding handler
        handler_map = StacGeneratorFactory.match_handler(source_configs)
        result: list[Any] = []
        for k, v in handler_map.items():
            for item in v:
                if config_type == "json":
                    result.append(k.create_config(item))
                else:
                    result.append(pd.DataFrame([k.create_config(item)]))
        # Generate config template with pre-filled band/column info
        match config_type:
            case "json":
                with Path(dst).open("w") as file:
                    json.dump(result, file)
            case "csv":
                df = pd.concat(cast(Iterable[pd.DataFrame], result))
                if "column_info" in df.columns:
                    df["column_info"] = df["column_info"].apply(lambda item: json.dumps(item))
                if "band_info" in df.columns:
                    df["band_info"] = df["band_info"].apply(lambda item: json.dumps(item))
                df.to_csv(dst, index=False)

    @staticmethod
    def match_handler(source_configs: list[str]) -> dict[type[ItemGenerator], list[dict[str, Any]]]:
        configs: list[dict[str, Any]] = []
        for source_config in source_configs:
            configs.extend(read_source_config(source_config))
        handler_map: dict[type[ItemGenerator], list[dict[str, Any]]] = collections.defaultdict(list)
        for config in configs:
            base_config = SourceConfig(**config)
            handler = StacGeneratorFactory.get_handler(base_config.source_extension)
            handler_map[handler].append(config)
        return handler_map



import json
from typing import Any

from pydantic import BaseModel, Field, field_validator, model_validator

from stac_generator.core.base.schema import SourceConfig


class BandInfo(BaseModel):
    """Band information for raster data"""

    name: str
    common_name: str
    wavelength: str | int | float | None = Field(default=None)  # Can be float or None
    nodata: float | None = Field(default=0)  # Default nodata value
    data_type: str | None = Field(default="uint16")  # Default data type for raster band
    description: str | None = Field(default=None)

    @model_validator(mode="before")
    @classmethod
    def check_common_name(cls, data: Any) -> Any:
        # Common name is derived from name if not provided. If provided, ignore
        if (
            isinstance(data, dict)
            and data.get("common_name", None) is None
            and data.get("name", None) is not None
        ):
            data["common_name"] = data["name"]
        return data

    @field_validator("wavelength", mode="before")
    @classmethod
    def parse_wavelength(cls, v: str | float | None) -> float | None:
        if isinstance(v, float) or v is None:
            return v
        if isinstance(v, str) and v == "no band specified":
            return None
        raise ValueError("Invalid wavelength value: {v}")


class RasterConfig(SourceConfig):
    """Configuration for raster data sources"""

    band_info: list[BandInfo]
    """List of band information - REQUIRED"""
    epsg: int | None = None
    """EPSG code for the raster's coordinate reference system"""

    @field_validator("band_info", mode="before")
    @classmethod
    def parse_bands(cls, v: str | list) -> list[BandInfo]:
        if isinstance(v, list):
            return v
        if isinstance(v, str):
            parsed = json.loads(v)
            if not isinstance(parsed, list):
                raise ValueError("bands parameter expects a json serialisation of a lis of Band")
            return parsed
        raise ValueError(f"Invalid bands dtype: {type(v)}")



from collections.abc import Sequence
from pathlib import Path

import pandas as pd
import pystac
import pytest

from stac_generator.core.base.utils import read_source_config
from stac_generator.core.vector.generator import VectorGenerator

CONFIG_PATH = Path("tests/files/unit_tests/vectors/configs")


def load_items(file: str) -> Sequence[pystac.Item]:
    config_path = CONFIG_PATH / file
    config = read_source_config(str(config_path))
    generator = VectorGenerator(config)
    return generator.create_items()


def load_item(file: str) -> pystac.Item:
    return load_items(file)[0]


def test_given_invalid_wrong_layer_expects_raises() -> None:
    with pytest.raises(Exception):
        load_items("invalid_wrong_layer.json")


def test_given_invalid_epsg_expects_raises() -> None:
    with pytest.raises(ValueError):
        load_items("invalid_wrong_epsg.json")


def test_given_invalid_column_info_expects_raises() -> None:
    with pytest.raises(ValueError):
        load_items("invalid_column_info.json")


def test_given_no_column_info_expects_no_value_in_property() -> None:
    item = load_item("no_column_info.json")
    assert "column_info" not in item.properties


def test_given_column_info_expects_column_info_in_property() -> None:
    item = load_item("with_column_info.json")
    assert "column_info" in item.properties


def test_given_no_epsg_expects_epsg_automatically_generated_in_property() -> None:
    item = load_item("no_epsg.json")
    assert "proj:code" in item.properties
    assert item.properties["proj:code"] == "EPSG:1168"
    assert "proj:wkt2" in item.properties


def test_given_with_epsg_expects_epsg_in_property() -> None:
    item = load_item("with_epsg.json")
    assert "proj:code" in item.properties
    assert item.properties["proj:code"] == "EPSG:1168"


def test_given_layers_info_expects_multiple_layers() -> None:
    items = load_items("with_layer.json")
    assert len(items) == 2
    assert items[0].id == "Sunbury"
    assert items[1].id == "Werribee"


def test_given_join_file_invalid_no_join_field_expects_throw() -> None:
    with pytest.raises(ValueError):
        load_item("join_invalid_config_no_join_field.json")


def test_given_join_file_invalid_wrong_join_field_expects_throw() -> None:
    with pytest.raises(ValueError):
        load_item("join_invalid_config_wrong_join_field.json")


def test_given_join_file_invalid_no_join_attribute_expects_throw() -> None:
    with pytest.raises(ValueError):
        load_item("join_invalid_config_no_join_attribute.json")


def test_given_join_file_invalid_wrong_join_attribute_expects_throw() -> None:
    with pytest.raises(ValueError):
        load_item("join_invalid_config_wrong_join_attribute.json")


def test_given_join_file_invalid_no_join_column_info_expects_throw() -> None:
    with pytest.raises(ValueError):
        load_item("join_invalid_config_no_join_column_info.json")


def test_given_join_file_invalid_wrong_join_column_info_expects_throw() -> None:
    with pytest.raises(ValueError):
        load_item("join_invalid_config_wrong_join_column_info.json")


def test_given_join_file_invalid_wrong_join_date_column_expects_throw() -> None:
    with pytest.raises(ValueError):
        load_item("join_invalid_config_wrong_join_date_column.json")


def test_given_join_with_date_expects_correct_start_end_datetime() -> None:
    item = load_item("join_with_date.json")
    assert "column_info" in item.properties
    assert item.properties["column_info"] == [{"name": "Suburb_Name", "description": "Suburb_Name"}]
    assert "join_column_info" in item.properties
    assert item.properties["join_column_info"] == [
        {"name": "Sell_Price", "description": "Median Sales Price in 2025"},
        {"name": "Rent_Price", "description": "Median Rental Price in 2025"},
        {
            "name": "Sell/Rent",
            "description": "Ratio of Sales Price (in $1000) over Rental Price (in $)",
        },
    ]
    assert item.properties["join_file"] == "tests/files/unit_tests/vectors/price.csv"
    assert item.properties["join_field"] == "Area"
    assert item.properties["join_attribute_vector"] == "Suburb_Name"
    assert item.properties["join_T_column"] == "Date"
    assert item.properties["start_datetime"] == "2020-01-01T00:00:00Z"
    assert item.properties["end_datetime"] == "2025-01-01T00:00:00Z"


def test_given_join_with_no_date_expects_same_start_end_datetime() -> None:
    item = load_item("join_no_date.json")
    assert item.properties["column_info"] == [{"name": "Suburb_Name", "description": "Suburb_Name"}]
    assert item.properties["join_column_info"] == [
        {"name": "Distance", "description": "Driving Distance to CBD in km"},
        {
            "name": "Public_Transport",
            "description": "Time taken to reach CBD by public transport in minutes",
        },
        {"name": "Drive", "description": "Time taken to reach CBD by driving in minutes"},
        {"name": "Growth", "description": "Average 5 year growth in percentage in 2025"},
        {"name": "Yield", "description": "Average rental yield in 2025"},
    ]
    assert item.properties["join_file"] == "tests/files/unit_tests/vectors/distance.csv"
    assert item.properties["join_field"] == "Area"
    assert item.properties["join_attribute_vector"] == "Suburb_Name"
    assert item.properties["start_datetime"] == item.properties["end_datetime"]
    assert item.datetime == pd.Timestamp(item.properties["start_datetime"])



import json
from typing import Self

from pydantic import field_validator, model_validator

from stac_generator.core.base.schema import ColumnInfo, HasColumnInfo, SourceConfig


class VectorConfig(SourceConfig, HasColumnInfo):
    """Extended source config with EPSG code."""

    epsg: int | None = None
    """EPSG code for checking against EPSG code of the vector data"""

    layer: str | None = None
    """Vector layer for multi-layer shapefile"""

    join_file: str | None = None
    """Name of join csv file"""

    join_attribute_vector: str | None = None
    """Attribute of shapefile to join with text file"""

    join_field: str | None = None
    """Field from text file to join with vector file"""

    date_format: str | None = "ISO8601"
    """Date format of text file"""

    join_T_column: str | None = None
    """Time column of text file if any"""

    join_column_info: list[ColumnInfo] | list[str] | None = None
    """List of attributes associated with point/vector data"""

    @field_validator("join_column_info", mode="before")
    @classmethod
    def coerce_to_object_text(
        cls, v: str | list[str] | None
    ) -> list[str] | list[ColumnInfo] | None:
        """Convert json serialised string of column info into matched object"""
        if v is None:
            return None
        if isinstance(v, list):
            return v
        parsed = json.loads(v)
        if not isinstance(parsed, list):
            raise ValueError(
                "column_info field expects a json serialisation of a list of ColumnInfo or a list of string"
            )
        return parsed

    @model_validator(mode="after")
    def validate_join_data(self) -> Self:
        if self.join_file:
            if not self.join_attribute_vector:
                raise ValueError(
                    "join_attribute_vector field is expected when join_file is provided"
                )
            if not self.join_field:
                raise ValueError("join_field is expected when join_file is provided")
            if not self.join_column_info:
                raise ValueError("join_column_info is expected when join_file is provided")
        return self
